/**
 * Copyright (C) 2011 Angelo Zerr <angelo.zerr@gmail.com> and Pascal Leclercq <pascal.leclercq@gmail.com>
 *
 * All rights reserved.
 *
 * Permission is hereby granted, free  of charge, to any person obtaining
 * a  copy  of this  software  and  associated  documentation files  (the
 * "Software"), to  deal in  the Software without  restriction, including
 * without limitation  the rights to  use, copy, modify,  merge, publish,
 * distribute,  sublicense, and/or sell  copies of  the Software,  and to
 * permit persons to whom the Software  is furnished to do so, subject to
 * the following conditions:
 *
 * The  above  copyright  notice  and  this permission  notice  shall  be
 * included in all copies or substantial portions of the Software.
 *
 * THE  SOFTWARE IS  PROVIDED  "AS  IS", WITHOUT  WARRANTY  OF ANY  KIND,
 * EXPRESS OR  IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE  WARRANTIES OF
 * MERCHANTABILITY,    FITNESS    FOR    A   PARTICULAR    PURPOSE    AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE,  ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package fr.opensagres.xdocreport.document.preprocessor;

import java.io.IOException;
import java.io.Reader;
import java.io.StringWriter;
import java.io.Writer;

import org.apache.commons.io.IOUtils;

import fr.opensagres.xdocreport.core.XDocReportException;
import fr.opensagres.xdocreport.core.io.MultiWriter;
import fr.opensagres.xdocreport.core.io.StreamCancelable;
import fr.opensagres.xdocreport.core.io.XDocArchive;
import fr.opensagres.xdocreport.document.debug.IPreProcessorDebugger;
import fr.opensagres.xdocreport.template.formatter.FieldsMetadata;
import fr.opensagres.xdocreport.template.formatter.IDocumentFormatter;

/**
 * Abstract class for {@link IXDocPreprocessor}.
 * 
 */
public abstract class AbstractXDocPreprocessor implements IXDocPreprocessor {

	public void preprocess(String entryName, XDocArchive documentArchive,
			FieldsMetadata fieldsMetadata, IDocumentFormatter formater,
			IPreProcessorDebugger debugger) throws XDocReportException,
			IOException {
		if (fieldsMetadata == null) {
			fieldsMetadata = FieldsMetadata.EMPTY;
		}
		// 1) Start preprocess
		long startTime = -1;
		if (isDebuggable(debugger)) {
			// Debug start preprocess
			startTime = System.currentTimeMillis();
			debugger.startPreprocess(entryName);
		}

		// 2) Do preprocess
		Reader reader = null;
		Writer writer = null;
		boolean result = false;
		try {
			// Get reader + writer for the entry name.
			reader = documentArchive.getEntryReader(entryName);
			writer = getWriter(entryName, documentArchive, debugger);

			// Do preprocess which use reader and store teh result of preprocess
			// in the writer.
			result = preprocess(entryName, reader, writer, null,
					fieldsMetadata, formater);

			if (isDebuggable(debugger)) {
				if (debugger.isDisplayResult()) {
					// Debug display reult of preprocess
					debugger.resultOfPreprocess(entryName,
							((MultiWriter) writer).getWriter(1).toString());
				}
				// Debug end preprocess
				debugger.endPreprocess(entryName, System.currentTimeMillis()
						- startTime, null);
			}
		} catch (Throwable e) {
			if (isDebuggable(debugger)) {
				// Debug end processs with error
				debugger.endPreprocess(entryName, System.currentTimeMillis()
						- startTime, e);
			}
			if (e instanceof RuntimeException) {
				throw (RuntimeException) e;
			}
			if (e instanceof IOException) {
				throw (IOException) e;
			}
			if (e instanceof XDocReportException) {
				throw (XDocReportException) e;
			}
			throw new XDocReportException(e);
		} finally {
			// Close reader + writer
			if (reader != null) {
				IOUtils.closeQuietly(reader);
			}
			if (writer != null) {
				if (result) {
					IOUtils.closeQuietly(writer);
				} else {
					// preprocess was not done, don't modify the entry.
					((StreamCancelable) writer).cancel();
				}
			}
		}

	}

	private Writer getWriter(String entryName, XDocArchive documentArchive,
			IPreProcessorDebugger debugger) {
		if (isDebuggable(debugger) && debugger.isDisplayResult()) {
			return new MultiWriter(documentArchive.getEntryWriter(entryName),
					new StringWriter());
		}
		return documentArchive.getEntryWriter(entryName);
	}

	protected boolean isDebuggable(IPreProcessorDebugger debugger) {
		return debugger != null;
	}

	public abstract boolean preprocess(String entryName, Reader reader,
			Writer writer, Writer debugWriter, FieldsMetadata fieldsMetadata,
			IDocumentFormatter formater) throws XDocReportException, IOException;

}
