/**
 * Copyright (C) 2011 Angelo Zerr <angelo.zerr@gmail.com> and Pascal Leclercq <pascal.leclercq@gmail.com>
 *
 * All rights reserved.
 *
 * Permission is hereby granted, free  of charge, to any person obtaining
 * a  copy  of this  software  and  associated  documentation files  (the
 * "Software"), to  deal in  the Software without  restriction, including
 * without limitation  the rights to  use, copy, modify,  merge, publish,
 * distribute,  sublicense, and/or sell  copies of  the Software,  and to
 * permit persons to whom the Software  is furnished to do so, subject to
 * the following conditions:
 *
 * The  above  copyright  notice  and  this permission  notice  shall  be
 * included in all copies or substantial portions of the Software.
 *
 * THE  SOFTWARE IS  PROVIDED  "AS  IS", WITHOUT  WARRANTY  OF ANY  KIND,
 * EXPRESS OR  IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE  WARRANTIES OF
 * MERCHANTABILITY,    FITNESS    FOR    A   PARTICULAR    PURPOSE    AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE,  ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package fr.opensagres.xdocreport.document.preprocessor.sax;

import java.util.Collection;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

import fr.opensagres.xdocreport.core.utils.StringUtils;
import fr.opensagres.xdocreport.template.formatter.FieldsMetadata;
import fr.opensagres.xdocreport.template.formatter.IDocumentFormatter;

public abstract class TransformedBufferedDocumentContentHandler extends
		BufferedDocumentContentHandler {

	private final FieldsMetadata fieldsMetadata;
	private final IDocumentFormatter formatter;
	private TableBufferedRegion currentTable;
	private RowBufferedRegion currentRow;

	protected TransformedBufferedDocumentContentHandler(
			FieldsMetadata fieldsMetadata, IDocumentFormatter formater) {
		this.fieldsMetadata = fieldsMetadata;
		this.formatter = formater;
	}

	@Override
	public void startDocument() throws SAXException {
		String directive = formatter != null ? formatter
				.getStartDocumentDirective() : null;
		if (StringUtils.isNotEmpty(directive)) {
			this.bufferedDocument.append(directive);
		}
		super.startDocument();
	}

	@Override
	public void endDocument() throws SAXException {
		String directive = formatter != null ? formatter
				.getEndDocumentDirective() : null;
		if (StringUtils.isNotEmpty(directive)) {
			this.bufferedDocument.append(directive);
		}
		super.endDocument();
	}

	@Override
	public boolean doStartElement(String uri, String localName, String name,
			Attributes attributes) throws SAXException {
		if (isTable(uri, localName, name)) {
			currentTable = new TableBufferedRegion();
			currentRegion.addRegion(currentTable);
			currentRegion = currentTable;

		} else if (isRow(uri, localName, name)) {
			// Check if currentRow belong to a table
			if (currentTable == null) {
				throw new SAXException(
						"XML mal formatted. XML Row must be included in a XML Table");
			}
			currentRow = new RowBufferedRegion();
			currentRegion.addRegion(currentRow);
			currentRegion = currentRow;
		}
		return super.doStartElement(uri, localName, name, attributes);
	}

	@Override
	public void doEndElement(String uri, String localName, String name)
			throws SAXException {
		super.doEndElement(uri, localName, name);
		if (isTable(uri, localName, name)) {
			currentRegion = bufferedDocument;
		} else if (isRow(uri, localName, name)) {
			currentRegion = currentTable;
		}
	}

	@Override
	protected void flushCharacters(String characters) {
		super.flushCharacters(processRowIfNeeded(characters));
	}

	protected String processRowIfNeeded(String content) {
		if (currentRow != null && formatter != null) {
			// characters parsing belong to a row
			// search if it contains fields list from metadata
			Collection<String> fieldsAsList = fieldsMetadata.getFieldsAsList();
			if (!currentRow.isLoopTemplateDirectiveInitilalized()) {
				for (final String fieldName : fieldsAsList) {
					if (content.contains(fieldName)) {
						String itemNameList = formatter.extractItemNameList(
								content, fieldName);
						if (StringUtils.isNotEmpty(itemNameList)) {
							currentRow.initializeLoopTemplateDirective(
									itemNameList, formatter);
							break;
						}
					}
				}
			}

			if (currentRow.isLoopTemplateDirectiveInitilalized()) {
				for (final String fieldName : fieldsAsList) {
					if (content.contains(fieldName)) {
						String newContent = formatter.formatAsFieldItemList(
								content, fieldName);
						if (newContent != null) {
							return newContent;
						}
					}
				}
			}
		}
		return content;
	}

	protected abstract boolean isTable(String uri, String localName, String name);

	protected abstract boolean isRow(String uri, String localName, String name);

}
